// ****************************************************************
//
//   IBM Confidential
//   OCO Source Materials
//   <<PART NUMBER - 5724-B75>>
//   (C) Copyright IBM Corp. 2001 - All Rights Reserved.
//   US Government Users Restricted Rights - Use, duplication or disclosure
//   restricted by GSA ADP Schedule Contract with IBM Corp.
//
// ****************************************************************

package com.ibm.etools.marshall.util;

import java.math.*;

public class MarshallExternalDecimalUtils
{
	private static final String copyright = "(c) Copyright IBM Corporation 2001,2003.";


// External Decimal Constants

// External Decimal Sign Value
public final static int EXTERNAL_DECIMAL_SIGN_EBCDIC = 0;
public final static int EXTERNAL_DECIMAL_SIGN_EBCDIC_CUSTOM = 1;
public final static int EXTERNAL_DECIMAL_SIGN_ASCII = 2;

// External Decimal Sign Format
public final static int SIGN_FORMAT_LEADING = 0;
public final static int SIGN_FORMAT_TRAILING = 1;
public final static int SIGN_FORMAT_LEADING_SEPARATE = 2;
public final static int SIGN_FORMAT_TRAILING_SEPARATE = 3;


static public void marshallExternalDecimalIntoBuffer (BigDecimal value, byte[] buffer, int startingOffset, int size, boolean signed, int virtualDecimalPoint, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	marshallExternalDecimalIntoBuffer (value, buffer, startingOffset, size, signed, virtualDecimalPoint, signFormat, externalDecimalSign, null);
	return;
}

static public void marshallExternalDecimalIntoBuffer (BigDecimal value, byte[] buffer, int startingOffset, int size, boolean signed, int virtualDecimalPoint, int signFormat, int externalDecimalSign, String codePage)
	throws IllegalArgumentException
{
	// Move the decimal point to the right
	value = value.movePointRight (virtualDecimalPoint);

	// !!! Need to make a decision here as to whether the value can fit into a long (8 bytes)
	// or not and then marshall appropriately (ie. 32 bit or 64 bit)
	if (size < 19)
		marshallExternalDecimalIntoBuffer (value.longValue(), buffer, startingOffset, size, signed, signFormat, externalDecimalSign, codePage);
	else
	{
		char[] extdChars = (value.toString()).toCharArray();
		byte[] extdBytes = ConversionUtils.convertToExternalDecimal (extdChars, size, signed, signFormat, externalDecimalSign, codePage);
		System.arraycopy (extdBytes, 0, buffer, startingOffset, extdBytes.length);
	}

	return;
}

static public void marshallExternalDecimalIntoBuffer (BigInteger value, byte[] buffer, int startingOffset, int size, boolean signed, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	marshallExternalDecimalIntoBuffer (value, buffer, startingOffset, size, signed, signFormat, externalDecimalSign, null);
	return;
}

static public void marshallExternalDecimalIntoBuffer (BigInteger value, byte[] buffer, int startingOffset, int size, boolean signed, int signFormat, int externalDecimalSign, String codePage)
	throws IllegalArgumentException
{
	BigDecimal bd = new BigDecimal(value);

	marshallExternalDecimalIntoBuffer (bd, buffer, startingOffset, size, signed, 0, signFormat, externalDecimalSign, codePage);

	return;
}

static public void marshallExternalDecimalIntoBuffer (double value, byte[] buffer, int startingOffset, int size, boolean signed, int virtualDecimalPoint, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	marshallExternalDecimalIntoBuffer (value, buffer, startingOffset, size, signed, virtualDecimalPoint, signFormat, externalDecimalSign, null);
	return;
}

static public void marshallExternalDecimalIntoBuffer (double value, byte[] buffer, int startingOffset, int size, boolean signed, int virtualDecimalPoint, int signFormat, int externalDecimalSign, String codePage)
	throws IllegalArgumentException
{
	BigDecimal bd = new BigDecimal(Double.toString(value));

	marshallExternalDecimalIntoBuffer (bd, buffer, startingOffset, size, signed, virtualDecimalPoint, signFormat, externalDecimalSign, codePage);

	return;
}

static public void marshallExternalDecimalIntoBuffer (float value, byte[] buffer, int startingOffset, int size, boolean signed, int virtualDecimalPoint, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	marshallExternalDecimalIntoBuffer (value, buffer, startingOffset, size, signed, virtualDecimalPoint, signFormat, externalDecimalSign, null);
	return;
}

static public void marshallExternalDecimalIntoBuffer (float value, byte[] buffer, int startingOffset, int size, boolean signed, int virtualDecimalPoint, int signFormat, int externalDecimalSign, String codePage)
	throws IllegalArgumentException
{
	BigDecimal bd = new BigDecimal(Float.toString(value));

	marshallExternalDecimalIntoBuffer (bd, buffer, startingOffset, size, signed, virtualDecimalPoint, signFormat, externalDecimalSign, codePage);

	return;
}

static public void marshallExternalDecimalIntoBuffer (long value, byte[] buffer, int startingOffset, int size, boolean signed, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	marshallExternalDecimalIntoBuffer (value, buffer, startingOffset, size, signed, signFormat, externalDecimalSign, null);
	return;
}

static public void marshallExternalDecimalIntoBuffer (long value, byte[] buffer, int startingOffset, int size, boolean signed, int signFormat, int externalDecimalSign, String codePage)
	throws IllegalArgumentException
{
	byte[] pdBytes = ConversionUtils.convertToExternalDecimal (value, size, signed, signFormat, externalDecimalSign, codePage);
	System.arraycopy (pdBytes, 0, buffer, startingOffset, pdBytes.length);

	return;
}

static public byte unmarshallByteFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	return (unmarshallByteFromBuffer (buffer, startingOffset, size, signed, signFormat, externalDecimalSign, null));
}

static public byte unmarshallByteFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int signFormat, int externalDecimalSign, String codePage)
	throws IllegalArgumentException
{
	byte result = (byte)unmarshallLongFromBuffer (buffer, startingOffset, size, signed, signFormat, externalDecimalSign, codePage);

	return (result);
}

static public short unmarshallShortFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	return (unmarshallShortFromBuffer (buffer, startingOffset, size, signed, signFormat, externalDecimalSign, null));
}

static public short unmarshallShortFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int signFormat, int externalDecimalSign, String codePage)
	throws IllegalArgumentException
{
	short result = (short)unmarshallLongFromBuffer (buffer, startingOffset, size, signed, signFormat, externalDecimalSign, codePage);

	return (result);
}

static public int unmarshallIntFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	return (unmarshallIntFromBuffer (buffer, startingOffset, size, signed, signFormat, externalDecimalSign, null));
}

static public int unmarshallIntFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int signFormat, int externalDecimalSign, String codePage)
	throws IllegalArgumentException
{
	int result = (int)unmarshallLongFromBuffer (buffer, startingOffset, size, signed, signFormat, externalDecimalSign, codePage);

	return (result);
}

static public long unmarshallLongFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	return (unmarshallLongFromBuffer (buffer, startingOffset, size, signed, signFormat, externalDecimalSign, null));
}

static public long unmarshallLongFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int signFormat, int externalDecimalSign, String codePage)
	throws IllegalArgumentException
{
	long result = 0;
	byte[] pdBytes = new byte[size];
	System.arraycopy (buffer, startingOffset, pdBytes, 0, pdBytes.length);	

	result = ConversionUtils.convertExternalDecimalToBinary (pdBytes, signed, signFormat, externalDecimalSign, codePage);

	return (result);
}

static public BigDecimal unmarshallBigDecimalFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int virtualDecimalPoint, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	return (unmarshallBigDecimalFromBuffer (buffer, startingOffset, size, signed, virtualDecimalPoint, signFormat, externalDecimalSign, null));
}

static public BigDecimal unmarshallBigDecimalFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int virtualDecimalPoint, int signFormat, int externalDecimalSign, String codePage)
	throws IllegalArgumentException
{
	BigDecimal bdRes = null;

	// !!! Need to make a decision here as to whether the value can fit into a long (8 bytes)
	// or not and then marshall appropriately (ie. 32 bit or 64 bit)
	if (size < 19)
	{
		long result = unmarshallLongFromBuffer (buffer, startingOffset, size, signed, signFormat, externalDecimalSign, codePage);

		// Create BigDecimal Resut
		bdRes = new BigDecimal(Long.toString(result));
	}
	else
	{
		byte[] extdBytes = new byte[size];
		System.arraycopy (buffer, startingOffset, extdBytes, 0, extdBytes.length);	

		bdRes = ConversionUtils.convertExternalDecimalToBigDecimal (extdBytes, signed, signFormat, externalDecimalSign, codePage);
	}

	// Move the decimal point left
	bdRes = bdRes.movePointLeft (virtualDecimalPoint);

	return (bdRes);
}

static public BigInteger unmarshallBigIntegerFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	return (unmarshallBigIntegerFromBuffer (buffer, startingOffset, size, signed, signFormat, externalDecimalSign, null));
}

static public BigInteger unmarshallBigIntegerFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int signFormat, int externalDecimalSign, String codePage)
	throws IllegalArgumentException
{
	BigInteger result = (unmarshallBigDecimalFromBuffer (buffer, startingOffset, size, signed, 0, signFormat, externalDecimalSign, codePage)).toBigInteger();

	return (result);
}

static public double unmarshallDoubleFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int virtualDecimalPoint, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	return (unmarshallDoubleFromBuffer (buffer, startingOffset, size, signed, virtualDecimalPoint, signFormat, externalDecimalSign, null));
}

static public double unmarshallDoubleFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int virtualDecimalPoint, int signFormat, int externalDecimalSign, String codePage)
	throws IllegalArgumentException
{
	double result = (unmarshallBigDecimalFromBuffer (buffer, startingOffset, size, signed, virtualDecimalPoint, signFormat, externalDecimalSign, codePage)).doubleValue();

	return (result);
}

static public float unmarshallFloatFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int virtualDecimalPoint, int signFormat, int externalDecimalSign)
	throws IllegalArgumentException
{
	return (unmarshallFloatFromBuffer (buffer, startingOffset, size, signed, virtualDecimalPoint, signFormat, externalDecimalSign, null));
}

static public float unmarshallFloatFromBuffer (byte[] buffer, int startingOffset, int size, boolean signed, int virtualDecimalPoint, int signFormat, int externalDecimalSign, String codePage)
	throws IllegalArgumentException
{
	float result = (unmarshallBigDecimalFromBuffer (buffer, startingOffset, size, signed, virtualDecimalPoint, signFormat, externalDecimalSign, codePage)).floatValue();

	return (result);
}


}

